/*-
 * Copyright (c) 2006 Bruno Ducrot <bruno@FreeBSD.org>
 * All rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include "opt_acpi.h"

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/sysctl.h>
#include <sys/types.h>

#include <contrib/dev/acpica/acpi.h>
#include <dev/acpica/acpivar.h>

#define _COMPONENT	ACPI_OEM
ACPI_MODULE_NAME("Dell")

ACPI_SERIAL_DECL(dell, "ACPI Dell extras");

static int			acpi_dell_hotkey;

static struct sysctl_ctx_list	acpi_dell_sysctl_ctx;
static struct sysctl_oid	*acpi_dell_sysctl_tree;


static int
acpi_dell_smi(int arg0, int arg1, int *ret)
{
	ACPI_STATUS status;
	ACPI_BUFFER buf;
	ACPI_OBJECT arg[2], obj;
	ACPI_OBJECT_LIST args;

	arg[0].Type = ACPI_TYPE_INTEGER;
	arg[0].Integer.Value = arg0;
	arg[1].Type = ACPI_TYPE_INTEGER;
	arg[1].Integer.Value = arg1;

	args.Count = 2;
	args.Pointer = arg;

	buf.Length = sizeof(obj);
	buf.Pointer = &obj;

	ACPI_SERIAL_BEGIN(dell);
	status = AcpiEvaluateObject(ACPI_ROOT_OBJECT, "SMI", &args, &buf);
	ACPI_SERIAL_END(dell);

	if (ACPI_SUCCESS(status) && obj.Type == ACPI_TYPE_INTEGER) {
		*ret = obj.Integer.Value;
		return (0);
	}

	printf("acpi_dell: can not call \\SMI()\n");
	return (ENXIO);
}

static int
acpi_dell_hotkey_sysctl(SYSCTL_HANDLER_ARGS)
{
	unsigned int val;

	acpi_dell_smi(0x25, 0, &acpi_dell_hotkey);

	val = *(u_int *)oidp->oid_arg1;
	return (sysctl_handle_int(oidp, &val, 0, req));
}

static int
acpi_dell_init(void)
{
	acpi_dell_sysctl_tree = SYSCTL_ADD_NODE(&acpi_dell_sysctl_ctx,
	    SYSCTL_STATIC_CHILDREN(_hw), OID_AUTO, "dell", CTLFLAG_RD, 0,
	        "ACPI Dell Extras");

	SYSCTL_ADD_PROC(&acpi_dell_sysctl_ctx,
	    SYSCTL_CHILDREN(acpi_dell_sysctl_tree), OID_AUTO,
	    "hotkey", CTLTYPE_INT | CTLFLAG_RD, &acpi_dell_hotkey, 0,
	    acpi_dell_hotkey_sysctl, "I", "Hotkey status");

	return (0);
}

static int
acpi_dell_loader(struct module *m, int what, void *arg)
{
	int rv = 0;
	int i;
	ACPI_HANDLE h;
	ACPI_STATUS status;
	ACPI_BUFFER vers_buf;
	ACPI_OBJECT *res;
	char buf[30];


	switch (what) {
	case MOD_LOAD:
		acpi_dell_sysctl_tree = NULL;
		sysctl_ctx_init(&acpi_dell_sysctl_ctx);
		vers_buf.Length = ACPI_ALLOCATE_BUFFER;
		vers_buf.Pointer = NULL;
		ACPI_SERIAL_BEGIN(dell);
		status = AcpiEvaluateObject(ACPI_ROOT_OBJECT, "VERS", NULL,
		    &vers_buf);
		ACPI_SERIAL_END(dell);
		if (ACPI_FAILURE(status)) {
			printf("can't evaluate %sVERS - %s\n",
			    acpi_name(ACPI_ROOT_OBJECT),
			    AcpiFormatException(status));
			rv = 0;
			break;
		}
		if (ACPI_FAILURE(AcpiGetHandle(ACPI_ROOT_OBJECT, "SMI", &h))) {
			printf("no SMI %p\n", &h);
			rv = ENXIO;
			break;
		}
		res = (ACPI_OBJECT *) vers_buf.Pointer;
		if (!ACPI_PKG_VALID(res, 1)) {
			printf("\\VERS is not a valid package\n");
			break;
		}
		printf("VERS count: %d\n", res->Package.Count);
		for (i = 0; i < res->Package.Count; ++i) {
			status = acpi_PkgStr(res, i, buf, 30);
			printf("[%s] %d\n", buf, status);
		}
		if (res)
			AcpiOsFree(res);
		rv = acpi_dell_init();
		break;
	case MOD_UNLOAD:
		sysctl_ctx_free(&acpi_dell_sysctl_ctx);
		acpi_dell_sysctl_tree = NULL;
		break;
	default:
		rv = EINVAL;
		break;
	}

	return (rv);
}

static moduledata_t acpi_dell_mod = {
	"acpi_dell",
	acpi_dell_loader,
	NULL
};

DECLARE_MODULE(acpi_dell, acpi_dell_mod, SI_SUB_KLD, SI_ORDER_ANY);
MODULE_DEPEND(acpi_video, acpi, 1, 1, 1);
